#ifndef _FIELD2D_H_
#define _FIELD2D_H_

#include <vector>

using std::vector;



template < class T >
class Field2D
{
public:
	typedef std::vector<T>::iterator iterator;
	typedef std::vector<T>::const_iterator const_iterator;

	Field2D( const vector<T>& Contents, unsigned int W, unsigned int H )
	:	m_Contents( Contents ),
		m_Width( W ),
		m_Height( H )
	{}

	Field2D( unsigned int W, unsigned int H )
	:	m_Width( W ),
		m_Height( H )
	{
		m_Contents.resize( W*H );
	}

	Field2D()
	:	m_Width( 0 ),
		m_Height( 0 )
	{}

	void Init( const vector<T>& Contents, unsigned int W, unsigned int H )
	{
		m_Contents = Contents;
		m_Width = W;
		m_Height = H;
	}

	void Init( unsigned int W, unsigned int H )
	{
		m_Contents.resize( W*H );
		m_Width = W;
		m_Height = H;
	}

	const T& Access( unsigned int i, unsigned int j ) const
	{
		return m_Contents[j*m_Width + i];
	}

	T& Access( unsigned int i, unsigned int j )
	{
		return m_Contents[j*m_Width + i];
	}

	inline void Fill( const T& Val )
	{
		m_Contents.assign( m_Width*m_Height, Val );
	}

	inline iterator begin() { return m_Contents.begin(); }
	inline iterator end() { return m_Contents.begin(); }

	inline const_iterator begin() const { return m_Contents.begin(); }
	inline const_iterator end() const { return m_Contents.begin(); }

	inline unsigned int Width() const { return m_Width; }
	inline unsigned int Height() const { return m_Height; }

private:

	vector<T>		m_Contents;
	unsigned int	m_Width;
	unsigned int	m_Height;

};


#endif //_FIELD2D_H_
