#include "lic.h"

#include <stdio.h>
#include <fstream>

using std::ofstream;



void LIC::LoadImage( const char* FileName )
{
	FILE* File;
	char Buffer[2000];
	int W, H;
	int Max;

	File = fopen( FileName, "r" );
	if( !File ) throw Exception( string( "Failed to open " ) + string( FileName ) + string( " for reading" ) );

	fgets( Buffer, 2000, File );
	if( strcmp( Buffer, "P2\n" ) ) throw Exception( string( FileName ) + string( " : not a valid pgm 2 file" ) );

	do
	{
		fgets( Buffer, 2000, File );
	}
	while( Buffer[0] == '#' );

	if( sscanf( Buffer, "%d %d\n", &W, &H ) != 2 ) throw Exception( string( FileName ) + string( " : failed to read dimensions" ) );
	if( fscanf( File, "%d\n", &Max ) != 1 ) throw Exception( string( FileName ) + string( " : failed to read max value" ) );

	m_ImageIn.Init( W, H );

	for( int j=0 ; j<H ; ++j )
	{
		for( int i=0 ; i<W ; ++i )
		{
			int Val;
			if( fscanf( File, "%d", &Val ) != 1 ) throw Exception( string( FileName ) + string( " : failed to read pixel value" ) );
			m_ImageIn.Access( i, j ) = (unsigned char)(Val*256/Max);
		}
		fscanf( File, "\n" );
	}

	m_Width = W;
	m_Height = H;

	fclose( File );

}

void LIC::LoadVectorField( const char* FileName )
{
	FILE* File;

	File = fopen( FileName, "rb" );
	if( !File ) throw Exception( string( "Failed to open " ) + string( FileName ) + string( " for reading" ) );

	float Wf,Hf;
	unsigned int W, H;

	fread( &Wf, sizeof( Wf ), 1, File );
	fread( &Hf, sizeof( Hf ), 1, File );

	W = (unsigned int) Wf;
	H = (unsigned int)Hf;

	m_VectorField.Init( W, H );

	for( int j=0 ; j<H ; ++j )
	{
		for( int i=0 ; i<W ; ++i )
		{
			float Vx, Vy;
			fread( &Vx, sizeof( Vx ), 1, File );
			fread( &Vy, sizeof( Vy ), 1, File );
			Vector3D V( Vx, Vy, 0 );
			/*if( fabs( Vx ) > 0.0001f && fabs( Vy ) > 0.0001f )*/ V.Normalize();
			m_VectorField.Access( i, j ) = V;
		}
	}

	m_Width = W;
	m_Height = H;

	fclose( File );
}

void LIC::LoadImage( const T_CharField& Image )
{
	m_ImageIn = Image;
}

void LIC::GenImage()
{
	m_ImageIn.Init( m_Width, m_Height );
	srand( time( NULL ) );
	for( int j=0 ; j<m_Height ; ++j )
	{
		for( int i=0 ; i<m_Width ; ++i )
		{
			m_ImageIn.Access( i, j ) = rand()%256;
		}
	}
}

void LIC::LoadVectorField( const T_VectorField& VectorField )
{
	m_VectorField = VectorField;
}

void LIC::SaveResult( const char* FileName )
{
	AdjustContrast();

	ofstream F( FileName );

	if( !F.good() ) throw Exception( string( "Failed to open " ) + string( FileName ) + string( " for writing" ) );

	F << "P2" << endl;
	F << m_Width << " " << m_Height << endl;
	F << 256 << endl;

	for( int j=0 ; j<m_Height ; ++j )
	{
		for( int i=0 ; i<m_Width ; ++i )
		{
			F << (int)(m_ImageOut.Access( i, j )) << " ";
		}
		F << endl;
	}
}

void LIC::AdjustContrast()
{
	int i,j;
	unsigned char Min = 255, Max = 0;

	for( j=0 ; j<m_Height ; ++j )
	{
		for( i=0 ; i<m_Width ; ++i )
		{
			unsigned char Val = m_ImageOut.Access( i, j );
			if( Val < Min ) Min = Val;
			if( Val > Max ) Max = Val;
		}
	}

	if( Min == 0 && Max == 255 ) return;

	unsigned char DMax = Max - Min;

	for( j=0 ; j<m_Height ; ++j )
	{
		for( i=0 ; i<m_Width ; ++i )
		{
			m_ImageOut.Access( i, j ) = (unsigned char)( (int)( m_ImageOut.Access( i, j ) - Min ) * 255 / DMax );
		}
	}
}

void LIC::DoLIC( int L )
{
	if( m_ImageIn.Width() == 0 ) GenImage();
	m_ImageOut.Init( m_Width, m_Height );

	for( int j=0 ; j<m_Height ; ++j )
	{
		for( int i=0 ; i<m_Width ; ++i )
		{
			Vector3D Position( i+.5f, j+.5f, 0 );
			float Integral( 0 );
			float Length( 0 );

			//Forward
			while( Length < L/2 )
			{
				int k = (int) Position.x(); if( Position.x() < 0 ) --k;
				int l = (int) Position.y(); if( Position.y() < 0 ) --l;
				if( k < 0 || l < 0 || k >= m_Width || l >= m_Height ) break;
				Vector3D Offset( Position.x() - k, Position.y() - l, 0 );
				Vector3D Direction = m_VectorField.Access( k, l );

				if( Direction.Len() < 0.1f ) break;

				float CP, A;

				CP = Direction.x();
				if( CP > 0.f )	A = 1 - Offset.x();
				else			A = Offset.x();
				float LRight = A/fabs( CP );

				CP = Direction.y();
				if( CP > 0.f )	A = 1 - Offset.y();
				else 			A = Offset.y();
				float LUp = A/fabs( CP );

				float d = min( LUp, LRight );
				//if( d <= 0.f ) cout << i << "," << j << ":" << Position << k << "," << l << endl;
				d += 0.01f;

				Integral += d * m_ImageIn.Access( k, l );
				Length += d;
				Position += Direction*d;
			}

			// Backward
			Position.Set( i+.5f, j+.5f, 0 );
			while( Length < L )
			{
				int k = (int) Position.x(); if( Position.x() < 0 ) --k;
				int l = (int) Position.y(); if( Position.y() < 0 ) --l;
				if( k < 0 || l < 0 || k >= m_Width || l >= m_Height ) break;
				Vector3D Offset( Position.x() - k, Position.y() - l, 0 );
				Vector3D Direction = m_VectorField.Access( k, l );
				Direction.Reverse();

				if( Direction.Len() < 0.1f ) break;

				float CP, A;

				CP = Direction.x();
				if( CP > 0.f )	A = 1 - Offset.x();
				else			A = Offset.x();
				float LRight = A/fabs( CP );

				CP = Direction.y();
				if( CP > 0.f )	A = 1 - Offset.y();
				else 			A = Offset.y();
				float LUp = A/fabs( CP );

				float d = min( LUp, LRight );
				//if( d <= 0.f ) cout << i << "," << j << ":" << Position << k << "," << l << endl;
				d += .01f;

				Integral += d * m_ImageIn.Access( k, l );
				Length += d;
				Position += Direction*d;
			}

			m_ImageOut.Access( i, j ) = (unsigned char) (Integral/Length);
		}
	}
}

float LIC::ComputeCurve( int x, int y, int MaxLen )
{
	float Sum = 0;

	if( m_VectorField.Access( x, y ).SLen() <= 0.1f ) return 0.f;

	int TLen = 0;
	for( int dir=0 ; dir<2 ; ++dir )
	{
		int i = x;
		int j = y;

		int Len = 0;
		while( Len < MaxLen )
		{
			Vector3D Position( i+.5f, j+.5f, 0 );
			Vector3D Direction = m_VectorField.Access( i, j );
			if( Direction.SLen() <= 0.1f ) break;
			//cout << Direction;
			Direction.Normalize();

			if( dir ) Position += Direction;
			else Position -= Direction;

			int a = int( Position.x() );
			int b = int( Position.y() );

			if( a<0 || b<0 || a>=m_Width || b>=m_Height ) break;

			Vector3D Direction2 = m_VectorField.Access( a, b );
			if( Direction2.SLen() <= 0.1f ) break;
			Direction2.Normalize();

			Sum += fabs( Direction * Direction2 );

			i = a;
			j = b;
			++Len;
		}
		TLen += Len;
		//cout << endl;
	}
	if( TLen > MaxLen / 4 )
	{
		Sum /= (float)TLen;
	}
	else Sum = 0;

	return Sum;
}

void LIC::DoDDA( int L )
{
	if( m_ImageIn.Width() == 0 ) GenImage();
	m_ImageOut.Init( m_Width, m_Height );
	m_Curve.Init( m_Width, m_Height );

	float Max = 0;
	float Min = FLT_MAX;

	for( int j=0 ; j<m_Height ; ++j )
	{
		for( int i=0 ; i<m_Width ; ++i )
		{
			float Val = ComputeCurve( i, j, L );
			if( Val > Max ) { Max = Val; cout << "Max : " << Max << endl; }
			if( Val < Min ) { Min = Val; cout << "Min : " << Min << endl; }
			m_Curve.Access( i, j ) = Val;
		}
	}

	for( int j=0 ; j<m_Height ; ++j )
	{
		for( int i=0 ; i<m_Width ; ++i )
		{
			float& Val = m_Curve.Access( i, j );
			Val = (Val-Min)/(Max-Min);
			m_ImageOut.Access( i, j ) = (unsigned char)( ( Val*Val*Val*Val ) * 255.f );
		}
	}

}




